package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.world.level.block.state.BlockState
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.energy.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.capability.matteryEnergy
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.nbt.map
import ru.dbotthepony.mc.otm.core.nbt.set

/**
 * Simple machine, which can work on only one job at time.
 *
 * From technical point, this is a specialized use case of [MachineJobEventLoop].
 */
abstract class MatteryWorkerBlockEntity<JobType : IMachineJob>(
	type: BlockEntityType<*>,
	blockPos: BlockPos,
	blockState: BlockState,
	val jobDeserializer: (tag: CompoundTag) -> JobType?
) : MatteryPoweredBlockEntity(type, blockPos, blockState) {
	val jobEventLoop = object : MachineJobEventLoop<JobType>() {
		override val energy: IMatteryEnergyStorage?
			get() = matteryEnergy
		override val isBlockedByRedstone: Boolean
			get() = redstoneControl.isBlockedByRedstone

		override fun deserializeJob(nbt: CompoundTag): JobType? {
			return jobDeserializer.invoke(nbt)
		}

		override fun onJobFinish(job: JobType): JobStatus {
			return this@MatteryWorkerBlockEntity.onJobFinish(job)
		}

		override fun computeNextJob(): JobContainer<JobType> {
			return this@MatteryWorkerBlockEntity.computeNextJob()
		}

		override fun jobUpdated(new: JobType?, old: JobType?) {
			this@MatteryWorkerBlockEntity.jobUpdated(new, old)
		}

		override fun onWorkTick(requiredPower: Decimal, extractedPower: Decimal, ticksAdvanced: Double, job: JobType): JobStatus {
			return this@MatteryWorkerBlockEntity.onWorkTick(requiredPower, extractedPower, ticksAdvanced, job)
		}
	}

	protected open fun jobUpdated(new: JobType?, old: JobType?) {}
	protected abstract fun onJobFinish(job: JobType): JobStatus
	protected abstract fun computeNextJob(): JobContainer<JobType>

	protected open fun onWorkTick(requiredPower: Decimal, extractedPower: Decimal, ticksAdvanced: Double, job: JobType): JobStatus {
		return JobStatus.SUCCESS
	}

	override fun saveShared(nbt: CompoundTag) {
		super.saveShared(nbt)
		nbt["JobLoop"] = jobEventLoop.serializeNBT()
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)
		nbt.map("JobLoop", jobEventLoop::deserializeNBT)
	}

	override fun setChanged() {
		super.setChanged()
		jobEventLoop.isIdling = false
	}

	override fun setChangedLight() {
		super.setChangedLight()
		jobEventLoop.isIdling = false
	}

	protected fun powerLevelUpdated() {
		super.setChangedLight()
		jobEventLoop.notify(MachineJobEventLoop.IdleReason.POWER)
	}

	protected fun itemContainerUpdated() {
		super.setChanged()
		jobEventLoop.notify(MachineJobEventLoop.IdleReason.ITEM)
	}

	protected fun matterLevelUpdated() {
		super.setChangedLight()
		jobEventLoop.notify(MachineJobEventLoop.IdleReason.MATTER)
	}

	override fun redstoneStatusUpdated(newBlocked: Boolean, oldBlocked: Boolean) {
		super.redstoneStatusUpdated(newBlocked, oldBlocked)
		jobEventLoop.isIdling = newBlocked
	}

	override fun tick() {
		super.tick()
		jobEventLoop.think()

		if (jobEventLoop.errorTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.ERROR) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.ERROR), Block.UPDATE_CLIENTS)
		} else if (jobEventLoop.workingTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.WORKING) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
		} else if (jobEventLoop.idleTicksAnim > 20 && blockState.hasProperty(WorkerState.WORKER_STATE) && blockState.getValue(WorkerState.WORKER_STATE) != WorkerState.IDLE) {
			level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
		}
	}

	companion object {
		fun appendHoverText(itemStack: ItemStack, blockGetter: BlockGetter?, tooltips: MutableList<Component>, flag: TooltipFlag) {
			itemStack.tag ?: return

			WorkerEnergyStorage.appendHoverText(itemStack, blockGetter, tooltips, flag)
			MatteryPoweredBlockEntity.appendHoverText(itemStack, blockGetter, tooltips, flag)
		}
	}
}
